from .. utils.do_jit import do_jit as __do_jit
from . jit_RK4 import jit_last_state as __jit_last_state

def last_state(RS, q0, t0, t_end, args = (), h = 1e-3):
    return __jit_last_state(__do_jit(RS), q0, t0, t_end, args, h)



from numba import njit

@njit
def __jit_stepRK4(RS, X, t_i, h, args):
    k1 = RS(X, t_i, *args)
    k2 = RS(X + h * k1 / 2, t_i + h / 2, *args)
    k3 = RS(X + h * k2 / 2, t_i + h / 2, *args)
    k4 = RS(X + h * k3, t_i + h, *args)
    return h * (k1 + 2 * k2 + 2 * k3 + k4) / 6

@njit
def jit_last_state(RS, q0, t0, t_end, args, h):
    '''
    Return last state at the t_end time moment.
    Take Right hand side, initial state, step by time,
    last time moment and args of system.
    '''
    X, t = q0.copy(), t0
    while t + h < t_end:
        X += __jit_stepRK4(RS, X, t, h, args)
        t += h

    X += __jit_stepRK4(RS, X, t, t_end - t, args)
    return X



from numba import njit
from numpy.linalg import inv, norm
import numpy as np

@njit
def __jit_yacoby_matrix_numeric(VF, X, args, h=1e-3):
    N = len(X)
    y_m = np.empty((N, N))
    X_1, X_2 = X.copy(), X.copy()

    for i in range(N):
        X_1[i] -= h
        X_2[i] += h
        y_m[:, i] = (VF(X_2, *args) - VF(X_1, *args)) / 2 / h
        X_1[i] += h
        X_2[i] -= h
    return y_m

@njit
def jit_newton_without_yacoby(VF, X0, eps, K_MAX, args, verify_x):
    X = X0
    Y_1 = inv(__jit_yacoby_matrix_numeric(VF, X, args))
    d = Y_1 @ VF(X, *args)
    iteration = 0
    while norm(d) > eps and iteration < K_MAX:
        if not verify_x(X):
            raise ArithmeticError("We can't find the point")
        iteration += 1
        Y_1 = inv(__jit_yacoby_matrix_numeric(VF, X, args))
        d = Y_1 @ VF(X, *args)
        X -= d
    if iteration == K_MAX:
        raise ArithmeticError("Too much iterations")
    return X

@njit
def jit_newton_with_yacoby(VF, X0, eps, K_MAX, args, jit_yacoby_matrix, verify_x):
    X = X0
    Y_1 = inv(jit_yacoby_matrix(X))
    d = Y_1 @ VF(X, *args)
    iteration = 0
    while norm(d) > eps and iteration < K_MAX:
        if not verify_x(X):
            raise ArithmeticError("We can't find the point")
        iteration += 1
        Y_1 = inv(jit_yacoby_matrix(X))
        d = Y_1 @ VF(X, *args)
        X -= d
    if iteration == K_MAX:
        raise ArithmeticError("Too much iterations")
    return X




import numpy as np
import math as mt
from ..optimizers import newton as __newton
from ..integrators import jit_RK4 as __jit_RK4

from ..utils.do_jit import do_jit as __do_jit

from .jit_main_method import vf_special as __vf_special
from .jit_main_method import super_rs as __super_rs
from .jit_main_method import verify_x as __verify_x 

def find_limit_cycle(RS, args, IC0, T0, phase_period = 2 * mt.pi, h=1e-3, eps=1e-3):
    '''TODO: Docs'''

    IC0 = IC0.copy()
    assert IC0[0] == 0 # Main Convention
    IC0[0] = T0
    # too slow :(
    # IC0 = __newton(lambda X: VF(X) ** 2, IC0, 1e-1, verify_x=__verify_x)
    IC = __newton(__vf_special, IC0, eps, args=(__do_jit(RS), args, phase_period, h), verify_x=__verify_x)
    T = IC[0]; IC[0] = 0
    return T, IC

def get_monogrommy_matrix(rs_orig, rs_linear,
                          q0_limit_cycle, T,
                          args_linear, args_orig,
                          h_integrate=1e-3):

    '''TODO: Docs'''
    RS_SIZE = len(q0_limit_cycle)
    SUPER_RS = __super_rs
    SUPER_ARGS = (__do_jit(rs_orig), __do_jit(rs_linear), args_linear, args_orig, RS_SIZE)

    M = np.empty((RS_SIZE, RS_SIZE))
    for i in range(RS_SIZE):
        q0_linear = np.zeros(RS_SIZE); q0_linear[i] = 1.0
        q0 = np.array([q0_linear, q0_limit_cycle]).flatten()
        M[i] = __jit_RK4.jit_last_state(SUPER_RS, q0, 0, T, SUPER_ARGS, h_integrate)[:RS_SIZE]
    return M




import numpy as np
import math as mt
from ..integrators import jit_RK4
from numba import njit

TIME_MAX = 5000

@njit
def verify_x(X):
    '''Check the time is valid'''
    return X[0] < TIME_MAX and X[0] > 0


@njit
def vf_special(X, RS, args, phase_period = 2 * mt.pi, h=1e-3):
    # Convention
    # X = [T, der0, ph1, der1, ph2, der2, ph3, der3, ..., phN, derN]
    # T = X[0]

    X = X.copy()

    T = X[0]; X[0] = 0
    last_state = jit_RK4.jit_last_state(RS, X, 0, T, args, h)

    phase_period_arr = np.zeros(len(X))
    phase_period_arr[::2] = phase_period

    return last_state - X - phase_period_arr

@njit
def super_rs(q, t, rs_orig, rs_linear, rs_linear_args, rs_orig_args, rs_size):
    # q = [ linear ... , orig ... ]
    q_linear = q[:rs_size]
    q_orig = q[rs_size:]
    rotate_phases = q_orig[::2] # phases

    lrs = rs_linear(q_linear, t, *rs_linear_args, rotate_phases)
    rs = rs_orig(q_orig, t, *rs_orig_args)

    final_rs = np.zeros(2 * rs_size)
    for i in range(rs_size):
        final_rs[i] = lrs[i]
        final_rs[i + rs_size] = rs[i]
    return final_rs



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rotary_states',
 'rotary_states.src',
 'rotary_states.src.integrators',
 'rotary_states.src.limit_cycles',
 'rotary_states.src.optimizers',
 'rotary_states.src.utils']

package_data = \
{'': ['*']}

install_requires = \
['numba>=0.55.0,<0.56.0', 'numpy>=1.18,<1.22', 'scipy>=1.7.3,<2.0.0']

setup_kwargs = {
    'name': 'rotary-states',
    'version': '0.1.1',
    'description': 'The minimalistic framework for finding rotational states',
    'long_description': None,
    'author': 'Dmitry Khorkin',
    'author_email': 'dmitryhorkin@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



from . jit_newton import jit_newton_with_yacoby as __jit_newton_with_yacoby, jit_newton_without_yacoby as __jit_newton_without_yacoby
from .. utils.do_jit import do_jit as __do_jit

def newton(VF, X0, eps=1e-3, K_MAX=100, yacoby_matrix = False, args = (), verify_x=lambda _: True):
    '''Explanation here https://www.wikiwand.com/en/Newton%27s_method'''
    
    if not yacoby_matrix:
        return __jit_newton_without_yacoby(__do_jit(VF), X0, eps, K_MAX, args, __do_jit(verify_x))

    return __jit_newton_with_yacoby(__do_jit(VF), X0, eps, K_MAX, args, __do_jit(yacoby_matrix), __do_jit(verify_x))
